#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.

"""
Useful functions
"""

# Python modules
import os
import re
from codecs import open
from collections import OrderedDict

# Third-party modules
from lxml import etree
import xmltodict
import requests

# Project modules
from fnapy.config import *
from fnapy.compat import to_unicode
from fnapy.exceptions import FnapyUpdateOfferError


# CLASSES

class Query(object):
    """A simple class to create queries"""
    def __init__(self, name, tags={}, **kwargs):
        self._dict = OrderedDict()
        self.name = name
        self.attrib = kwargs

        for k, v in kwargs.items():
            if not k.startswith('@'):
                self._dict['@'+k] = v

        for k, v in tags.items():
            self._dict[k] = v

    def _create_new_dict(self):
        new_dict = OrderedDict()
        for k, v in self._dict.items():
            if k.startswith('@'):
                new_dict[k] = v
        return new_dict

    def was(self, state):
        """Perform a query on the states

        :state: the name of a state

        Example
        >>> states = Query('state').was('Created')

        """
        new_dict = self._create_new_dict()
        new_dict['state'] = {'#text': state}
        return Query(self.name, tags=new_dict)

    def between(self, min, max):
        new_dict = self._create_new_dict()
        new_dict['min'] = {'#text': min}
        new_dict['max'] = {'#text': max}
        return Query(self.name, tags=new_dict)

    def _operator(self, op, value):
        op_to_mode = {'eq': 'Equals',
                      'ge': 'GreaterThanOrEquals', 'gt': 'GreaterThan',
                      'le': 'LessThanOrEquals', 'lt': 'LessThan'}
        new_dict = self._create_new_dict()
        new_dict['@mode'] = op_to_mode[op]
        new_dict['@value'] = value
        return Query(self.name, tags=new_dict)

    def eq(self, value):
        return self._operator('eq', value)

    def ge(self, value):
        return self._operator('ge', value)

    def gt(self, value):
        return self._operator('gt', value)

    def le(self, value):
        return self._operator('le', value)

    def lt(self, value):
        return self._operator('lt', value)

    @property
    def dict(self):
        """Return the OrderedDict"""
        return self._dict

    @property
    def xml(self):
        """Return the XML"""
        return dict2xml({self.name: self.dict})

    def __str__(self):
        return "<{0}: {1}>".format(self.__class__.__name__, self.name)

    __repr__ = __str__


class HttpMessage(object):
    def __init__(self, content):
        # we need to turn content into valid utf8
        content = to_unicode(content, encoding='iso-8859-1').encode('utf-8')


        # content is a string
        if len(content) != 0:
            self.dict = xml2dict(content)
            # etree._Element
            self.element = etree.fromstring(content)
        else:
            self.dict = OrderedDict()
            self.element = etree.Element('empty_content')

        # Raw XML
        self.xml = content

        self.tag = re.sub(pattern='{[^}]+}', repl='', string=self.element.tag,
                          flags=0)

    def __repr__(self):
        return '<{0}: {1}>'.format(self.__class__.__name__, self.tag)

    def __str__(self):
        return self.xml


class Request(HttpMessage):
    """A handy class to handle the request"""
    def __init__(self, content):
        super(Request, self).__init__(content)


class Response(HttpMessage):
    """A handy class to handle the response"""
    def __init__(self, content):
        super(Response, self).__init__(content)


# TODO Implement a check for the attributes
class Message(object):
    ACTIONS = (
        'mark_as_read', 'mark_as_unread', 'archive',
        'mark_as_read_and_archive', 'unarchive', 'reply', 'create'
    )

    TO = (
        'CALLCENTER', 'CLIENT', 'ALL'
    )

    SUBJECTS = (
        'product_information', 'shipping_information', 'order_information',
        'offer_problem', 'offer_not_received', 'other_question'
    )

    TYPES = ('ORDER', 'OFFER')

    def __init__(self, action, id, to='ALL', description='', subject='', type='ORDER'):
        self._action = action
        self._id = id
        self._to = to
        self._description = description
        self._subject = subject
        self._type = type

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, new_value):
        if new_value not in Message.ACTIONS:
            msg = 'Invalid action. Choose between {}.'.format(Message.ACTIONS)
            raise ValueError(msg)
        self._action = new_value

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, new_value):
        self._id = new_value

    @property
    def to(self):
        return self._to

    @to.setter
    def to(self, new_value):
        if new_value not in Message.TO:
            msg = 'Invalid recipient. Choose between {}.'.format(Message.TO)
            raise ValueError(msg)
        self._to = new_value

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, new_value):
        self._description = new_value

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, new_value):
        if new_value not in Message.SUBJECTS:
            msg = 'Invalid subject. Choose between {}'.format(Message.SUBJECTS)
            raise ValueError(msg)
        self._subject = new_value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, new_value):
        if new_value not in Message.TYPES:
            msg = 'Invalid type. Choose between {}'.format(Message.TYPES)
            raise ValueError(msg)
        self._type = new_value

    def __repr__(self):
        r = '<Message: action={self.action}, id={self.id}, to={self.to}, '
        r += 'description={self.description}, subject={self.subject}, '
        r += 'type={self.type}>'
        return r.format(self=self)

    def __str__(self):
        return """Message
action     : {self.action}
id         : {self.id}
to         : {self.to}
description: {self.description}
subject    : {self.subject}
type       : {self.type}
        """.format(self=self)

    def to_dict(self):
        """Return the a dictionary in the xmltodict format"""
        message = {'message': {
            '@action': self.action,
            '@id': self.id,
            'message_to': {'#text': self.to},
            'message_subject': {'#text': self.subject},
            'message_description': {'#text': self.description},
            'message_type': {'#text': self.type},
        }}
        return message


# FUNCTIONS
def get_url(sandbox=True):
    """Return the url for the sandbox or the real account

    Usage::
        url = get_url(sandbox=sandbox)

    :type sandbox: bool
    :param sandbox: determines whether you get the url for the sandbox
    account (True) or the real account (False).

    :rtype: str
    :returns: the entrypoint url to access the FNAC WebServices

    """
    use_sandbox = {True: "https://marketplace.ws.fd-recette.net/api.php/",
                   False: "https://vendeur.fnac.com/api.php/"}
    return use_sandbox[sandbox]


def get_credentials(sandbox=True):
    """Return the credentials for the sandbox or the real account

    Usage::
        credentials = get_credentials(sandbox=sandbox)

    :type sandbox: bool
    :param sandbox: determines whether you get the credentials for the sandbox
    account (True) or the real account (False).

    :rtype: dict
    :returns: the credentials for the selected account type

    """
    credentials = {
        'sandbox': {'partner_id': os.getenv('FNAC_SANDBOX_PARTNER_ID'),
                    'shop_id': os.getenv('FNAC_SANDBOX_SHOP_ID'),
                    'key': os.getenv('FNAC_SANDBOX_KEY')},
        'real': {'partner_id': os.getenv('FNAC_PARTNER_ID'),
                 'shop_id': os.getenv('FNAC_SHOP_ID'),
                 'key': os.getenv('FNAC_KEY')},
    }
    use_sandbox = {True: 'sandbox', False: 'real'}
    account_type = use_sandbox[sandbox]
    return credentials[account_type]


def dict2xml(_dict):
    """Returns a XML string from the input dictionary"""
    xml = xmltodict.unparse(_dict, pretty=True)
    xmlepured = remove_namespace(xml)
    return xmlepured


def remove_namespace(xml):
    """Remove the namespace from the XML string"""
    xmlepured = re.sub(pattern=' xmlns="[^"]+"', repl='', string=xml, flags=0)
    xmlepured = xmlepured.encode('utf-8')
    return xmlepured


def xpath(element, node_name):
    """A convenient function to look for nodes in the XML

    >>> nodes = xpath(response.element, node_name)

    """
    if '/' in node_name:
        node_name = '//ns:'.join(node_name.split('/'))
    return element.xpath('//ns:{0}'.format(node_name),
                         namespaces={'ns': XHTML_NAMESPACE})


def findall(element, node_name):
    """A convenient function to look for nodes in the XML

    >>> nodes = findall(response.element, node_name)

    """
    if '/' in node_name:
        node_name = '//ns:'.join(node_name.split('/'))
    return element.findall('.//ns:{0}'.format(node_name),
                         namespaces={'ns': XHTML_NAMESPACE})

def find(element, node_name):
    """Look for a node in the XML

    >>> node = find(response.element, node_name)

    """
    if '/' in node_name:
        node_name = '//ns:'.join(node_name.split('/'))
    return element.find('.//ns:{0}'.format(node_name),
                        namespaces={'ns': XHTML_NAMESPACE})


def extract_text(element, node_name):
    """Extract the text from the selected node

    If no text is found, the empty string is returned.

    """
    inner_element = find(element, node_name)
    return inner_element.text if inner_element is not None else ''


def xml2dict(xml):
    """Returns a dictionary from the input XML

    :type xml: unicode
    :param xml: The XML

    :rtype: dict
    :returns: the dictionary correspoding to the input XML
    """
    xmlepured = remove_namespace(to_unicode(xml))
    return xmltodict.parse(xmlepured)


def parse_xml(response, tag_name):
    """Get the text contained in the tag of the response

    :param response: the Response
    :param tag_name: the name of the tag
    :returns: the text enclosed in the tag

    """
    xml = etree.XML(response.content)
    elements = xml.xpath('//ns:{}'.format(tag_name),
                         namespaces={'ns': XHTML_NAMESPACE})
    return elements[0].text if len(elements) else ''


def check_offer_data(offer_data):
    """Check the offer_data passed to update_offers is valid

    :type offer_data: dict
    :param offer_data: the parameters used to update an offer

    :returns: None

    offer_data must be a dictionary with at least 2 keys:
    - offer_reference (the sku)
    - any other parameter allowed by the service (price, quantity,
    product_state, ...)

    Raises a FnapyUpdateOfferError if the offer_data is not valid.

    """
    if not isinstance(offer_data, dict):
        msg = 'The argument must be a dictionary.'
        raise FnapyUpdateOfferError(msg)

    if 'offer_reference' not in offer_data:
        msg = 'The dictionary must contain the key "offer_reference" (the sku)'
        raise FnapyUpdateOfferError(msg)

    valid_keys = [x.name for x in REQUEST_ELEMENTS['offers_update']]
    for key in offer_data:
        if key not in valid_keys:
            msg = '{0} is not a valid parameter for updating an offer. '
            msg += ' Choose amongst {1}'
            msg = msg.format(key, valid_keys)
            raise FnapyUpdateOfferError(msg)


# TODO Reimplement create_offer_element with kwargs
def create_offer_element(offer_data):
    """Create an offer element

    An offer needs at least one offer_reference (SKU) and any other parameter
    accepted by the service (cf documentation)
    :param offer_reference: a seller offer reference (such as SKU)
    :param product_reference: a product reference (such as EAN)
    :param product_state: a product state
    :param price: a price
    :param quantity: a quantity
    :param description: a description of the product

    :returns: offer (etree.Element)

    """
    offer = etree.Element('offer')
    offer_reference = offer_data['offer_reference']
    etree.SubElement(offer, "offer_reference", type="SellerSku").text = etree.CDATA(offer_reference)
    offer_data_items = [(k, v) for k, v in offer_data.items() if k != 'offer_reference']

    for key, value in offer_data_items:
        if key == 'product_reference':
            etree.SubElement(offer, 'product_reference', type="Ean").text = str(value)
        elif key == 'description':
            etree.SubElement(offer, 'description').text = etree.CDATA(value)
        else:
            etree.SubElement(offer, key).text = str(value)
    return offer


def create_xml_element(connection, token, name):
    """A helper function creating an etree.Element with the necessary
    attributes

    :param name: The name of the element
    :returns: etree.Element

    """
    return etree.Element(name, nsmap={None: XHTML_NAMESPACE},
                         shop_id=connection.shop_id,
                         partner_id=connection.partner_id, token=token)


def get_order_ids(orders_query_response):
    """Returns the order_ids in orders_query_response"""
    orders = orders_query_response.dict['orders_query_response'].get('order', None)
    order_ids = []
    if orders:
        if isinstance(orders, (list, tuple)):
            for order in orders:
                order_ids.append(order.get('order_id', ''))
        elif isinstance(orders, dict):
            order_ids.append(orders.get('order_id', ''))
    return order_ids


def get_token(sandbox=True):
    # We use the real seller account
    credentials = get_credentials(sandbox)
    partner_id = credentials['partner_id']
    shop_id = credentials['shop_id']
    key = credentials['key']

    xml = """<?xml version="1.0" encoding="utf-8"?>
<auth xmlns='http://www.fnac.com/schemas/mp-dialog.xsd'>
  <partner_id>{partner_id}</partner_id>
  <shop_id>{shop_id}</shop_id>
  <key>{key}</key>
</auth>
    """.format(partner_id=partner_id, shop_id=shop_id, key=key)
    url = get_url(sandbox)
    response = post(url, 'auth', xml)
    return parse_xml(response, 'token')


def set_credentials(xml, sandbox=True):
    """Set the credentials in the given raw XML """
    credentials = get_credentials(sandbox)
    creds = {'shop_id': credentials['shop_id'],
             'partner_id': credentials['partner_id'],
             'token': get_token(sandbox)}

    for cred, value in creds.items():
        xml = re.sub(pattern='{0}="[^"]+"'.format(cred),
                     repl='{0}="{1}"'.format(cred, value), string=xml, flags=0)
    return xml


def post(url, service, request):
    request = to_unicode(request).encode('utf-8')
    return requests.post(url + service, request, headers=HEADERS)


def save_xml_response(response, action):
    """Save the response in a file """
    output_dir = os.path.dirname(os.path.abspath(__file__))
    filename = action + '_response.xml'
    with open(os.path.join(output_dir, '../tests/assets', filename), 'w') as f:
        f.write(response.encode('utf-8'))
        print('Saved the response in {}'.format(filename))


def load_xml_request(action):
    input_dir = os.path.dirname(os.path.abspath(__file__))
    filename = action + '_request.xml'
    with open(os.path.join(input_dir, '../tests/assets', filename), 'r', 'utf-8') as f:
        request = f.read()
    return request



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.

"""

**fnapy** is a Python library using the FnacMarketPlace API to connect to your
own sales application to your FnacMarketplace seller account.  It uses the REST
WebService protocol to exchange data.

"""

# Python modules
from string import Template
import logging

# Third-party modules
import requests
from lxml import etree

# Project modules
from fnapy.utils import *
from fnapy.config import REQUEST_ELEMENTS, XHTML_NAMESPACE, HEADERS, XML_OPTIONS
from fnapy.compat import is_py3
from fnapy.connection import FnapyConnection
from fnapy.exceptions import FnapyUpdateOfferError


logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.setLevel(logging.DEBUG)
logger.addHandler(handler)


def _create_docstring(query_type):
    """Create the docstring for a given query method"""
    _query_docstring = Template(FnapyManager._query.__doc__)
    service = query_type + '_query'
    param_fmt = ':param {param.name}: {param.desc}'.format
    parameters = "\n\t".join(param_fmt(param=param) for param in REQUEST_ELEMENTS[service])
    return _query_docstring.substitute(query_type=query_type,
                                       parameters=parameters)


class FnapyManager(object):
    """A class to manage the different services provided by the FNAC API"""

    VALID_QUERY_TYPES = ('offers', 'orders', 'client_order_comments',
                         'messages', 'incidents', 'shop_invoices')

    def __init__(self, connection):
        """Initialize the manager"""
        if not isinstance(connection, FnapyConnection):
            raise TypeError('You must provide a valid FnapyConnection instance.')

        self.connection = connection
        self.auth_request = None
        self.offers_query_request = None
        self.offers_update_request = None
        self.orders_query_request = None
        self.batch_query_request = None
        self.batch_status_request = None
        self.orders_update_request = None
        self.carriers_query_request = None
        self.client_order_comments_query_request = None
        self.client_order_comments_update_request = None
        self.messages_query_request = None
        self.messages_update_request = None
        self.incidents_query_request = None
        self.incidents_update_request = None
        self.pricing_query_request = None
        self.shop_invoices_query_request = None

        # The batch_id updated every time an offer is updated
        self.batch_id = None

        # The url of the entrypoint
        self.url = get_url(self.connection.sandbox)

        # Authenticate when the manager is instanciated
        self.authenticate()

    def _get_response(self, element, xml):
        """Send the request and return the response as a dictionary

        :type element: lxml.etree.Element
        :param element: the XML element

        :type xml: str
        :param xml: the XML string sent in the request

        :returns: :class:`Response <Response>` object
        """
        service = element.tag
        response = requests.post(self.url + service, xml, headers=HEADERS)
        response = Response(response.content)
        if response.dict.get(service + '_response', {}).get('error'):
            # Reauthenticate and update the element
            element.attrib['token'] = self.authenticate()
            setattr(self, service + '_request',
                    Request(etree.tostring(element, **XML_OPTIONS)))
            # Resend the updated request
            response = requests.post(self.url + service,
                                     getattr(self, service + '_request').xml,
                                     headers=HEADERS)
            response = Response(response.content)
        return response

    def authenticate(self):
        """Authenticate to the FNAC API and return a token

        Usage::

            token = manager.authenticate()

        :returns: token
        :rtype: str

        """
        auth = etree.Element('auth', nsmap={None: XHTML_NAMESPACE})
        etree.SubElement(auth, 'partner_id').text = self.connection.partner_id
        etree.SubElement(auth, 'shop_id').text = self.connection.shop_id
        etree.SubElement(auth, 'key').text = self.connection.key
        self.auth_request = Request(etree.tostring(auth, **XML_OPTIONS))
        response = requests.post(self.url + 'auth', self.auth_request.xml,
                                 headers=HEADERS)
        self.token = parse_xml(response, 'token')
        return self.token

    def delete_offers(self, offer_references):
        """Delete the offers with the given offer_references (sku)

        Usage::

            response = manager.delete_offers(offer_references)

        :param offer_references: the list of SKUs corresponding to the offers
            you want to delete from your catalog

        :returns: :class:`Response <Response>` object

        """
        offers_update = create_xml_element(self.connection, self.token,
                                           'offers_update')
        for offer_reference in offer_references:
            offer = etree.Element('offer')
            etree.SubElement(offer, "offer_reference",
                             type="SellerSku").text = etree.CDATA(offer_reference)
            etree.SubElement(offer, 'treatment').text = 'delete'
            offers_update.append(offer)

        self.offers_update_request = Request(etree.tostring(offers_update,
                                                            **XML_OPTIONS))

        # the response contains the element batch_id
        response = self._get_response(offers_update,
                                      self.offers_update_request.xml)
        self.batch_id = response.dict['offers_update_response']['batch_id']
        return response

    # TODO Create a dictionary for the product_state
    def update_offers(self, offers_data):
        """Post the update offers and return the response

        Usage::

            response = manager.update_offers(offers_data)

        :type offers_data: list
        :param offers_data: the list of data to create the offers
                            where data is dictionary with the keys:

        * offer_reference  : the SKU (mandatory)
        * product_reference: the EAN (optional)
        * price            : the price of the offer (optional)
        * product_state    : an integer representing the state of the product
                             (documentation needed) (optional)
        * quantity         : the quantity (optional)
        * description      : a description of the offer (optional)

        The exception FnapyUpdateOfferError is raised if:
        - offer_reference and at least one of the optional parameters (except
        product_reference) are not provided
        - offers_data is empty

        :returns: :class:`Response <Response>` object

        """
        offers_update = create_xml_element(self.connection, self.token,
                                           'offers_update')

        if len(offers_data) == 0:
            msg = 'You must provide at least one offer_data.'
            raise FnapyUpdateOfferError(msg)

        for offer_data in offers_data:
            check_offer_data(offer_data)
            offer = create_offer_element(offer_data)
            offers_update.append(offer)
        self.offers_update_request = Request(etree.tostring(offers_update, **XML_OPTIONS))

        # the response contains the element batch_id
        response = self._get_response(offers_update, self.offers_update_request.xml)
        try:
            self.batch_id = response.dict['offers_update_response']['batch_id']
        except KeyError:
            self.batch_id = ''
        return response

    # TODO Improve the documentation
    def update_orders(self, order_id, order_update_action, actions):
        """Update the selected order with an order_update_action

        Usage::

            response = manager.update_orders(order_id, order_update_action, actions)

        :type order_id: str
        :param order_id: Order unique identifier from FNAC

        :type order_update_action: str
        :param order_update_action: Group action type for order detail action

        :type actions: list
        :param actions: a list of dictionaries with 2 keys:
            `'order_detail_id'` and `'action'`

        :returns: :class:`Response <Response>` object

        Available order_update_action:

        * accept_order       : The action for the order is accepting orders by the
                               seller
        * confirm_to_send    : The action for the order is confirming sending
                               orders by the seller
        * update             : The action for the order is updating orders by the seller
        * accept_all_orders  : The action for the order is accepting or refusing
                               all order_details of the order by the seller
        * confirm_all_to_send: The action for the order is confirming sending
                               all order_details by the seller
        * update_all         : The action for the order is to update tracking
                               information for all order_details

        Example: For this order (whose `order_id` is `'LDJEDEAS123'`), we have 2
        items. We decide to accept the first item and refuse the second::

            action1 = {"order_detail_id": 1, "action": "Accepted"}
            action2 = {"order_detail_id": 2, "action": "Refused"}
            response = manager.update_orders('LDJEDEAS123', 'accept_order', [action1, action2])

        """
        order_id = str(order_id)
        orders_update = create_xml_element(self.connection, self.token,
                                           'orders_update')
        order = etree.Element('order', order_id=order_id,
                              action=order_update_action)

        for action in actions:
            order_detail = etree.Element("order_detail")
            etree.SubElement(order_detail, 'order_detail_id').text = str(action['order_detail_id'])
            etree.SubElement(order_detail, 'action').text = str(action['action'])
            order.append(order_detail)

        orders_update.append(order)
        self.orders_update_request = Request(etree.tostring(orders_update, **XML_OPTIONS))
        return self._get_response(orders_update, self.orders_update_request.xml)

    # FIXME The batch_status_response doesn't contain the attributes (status)
    def get_batch_status(self, batch_id=None):
        """Return the status for the given batch id

        Usage::

            response = manager.get_batch_status(batch_id=batch_id)

        ..note:: :class:`FnapyManager <FnapyManager>` stores the last `batch_id`
                 but you can provide a new one if needed.

        :param batch_id: the batch id (optional)
        :returns: :class:`Response <Response>` object

        """
        if batch_id is not None:
            self.batch_id = batch_id

        batch_status = create_xml_element(self.connection, self.token, 'batch_status')
        etree.SubElement(batch_status, 'batch_id').text = self.batch_id
        self.batch_status_request = Request(etree.tostring(batch_status, **XML_OPTIONS))
        return self._get_response(batch_status, self.batch_status_request.xml)

    # TODO Implement this method that should handle any arguments to create
    # the xml properly
    def _check_elements(self, valid_elements, selected_elements):
        for element in selected_elements:
            if element not in (x.name for x in valid_elements):
                raise ValueError('{} is not a valid element.'.format(element))

    def _query(self, query_type, results_count='', **elements):
        """Query your catalog and return the ${query_type} response

        Usage::

            response = manager.query_${query_type}(results_count=results_count,
                                            **elements)

        The available XML elements are the following parameters:

        ${parameters}

        :returns: :class:`Response <Response>` object

        Examples: 
        Find the 2 first items  of the catalog::

            response = manager.query_${query_type}(results_count=2, paging=1)

        Find the ${query_type} created between 2 dates::

            >>> from fnapy.utils import Query
            >>> date = Query('date', type='Modified')
                .between(min="2016-08-23T17:00:00+00:00",
                         max="2016-08-26T17:00:00+00:00")
            >>> response = manager.query_${query_type}(date=date)

        """
        if query_type in FnapyManager.VALID_QUERY_TYPES:
            query_type += "_query"
        else:
            raise ValueError("The query_type must be in {}".format(FnapyManager.VALID_QUERY_TYPES))

        # TODO Refactor: Use a dictionary to prevent code duplication
        # Check the queried elements
        self._check_elements(REQUEST_ELEMENTS[query_type], elements.keys())

        # Make sure we have unicode
        # paging = str(paging).decode('utf-8')
        results_count = str(results_count)#.decode('utf-8')

        # Create the XML element
        query = create_xml_element(self.connection, self.token, query_type)
        if results_count:
            query.attrib['results_count'] = results_count

        # Create the XML from the queried elements 
        if len(elements):
            for key, value in elements.items():
                # Handle cases where Query is used
                if isinstance(value, Query):
                    value = value.dict
                d = {key: value}
                queried_elements = etree.XML(dict2xml(d))
                query.append(queried_elements)

        setattr(self, query_type + '_request', Request(etree.tostring(query, **XML_OPTIONS)))
        query_xml = getattr(self, query_type + '_request').xml
        return self._get_response(query, query_xml)

    # TODO generator for the paging
    # TODO Allow to specify the type of date ('Created', 'Modified'...)
    def query_offers(self, results_count='', **elements):
        return self._query('offers', results_count, **elements)

    def query_orders(self, results_count='', **elements):
        return self._query('orders', results_count, **elements)

    def query_pricing(self, eans):
        """Retrieve the best prices applied to a given product within all Fnac
        marketplace sellers (Fnac included)

        Usage::

            response = manager.query_pricing(eans)

        :type eans: list or tuple
        :param eans: a list of EANs

        :returns: response

        .. note: If no price is found for a product, a :class:`FnapyPricingError <FnapyPricingError>`
            is raised.

        """
        pricing_query = create_xml_element(self.connection, self.token, 'pricing_query')

        for ean in eans:
            product_reference = etree.Element("product_reference", type="Ean")
            product_reference.text = str(ean)
            pricing_query.append(product_reference)
        else:
            if len(eans) == 0:
                product_reference = etree.Element("product_reference", type="Ean")
                pricing_query.append(product_reference)

        self.pricing_query_request = Request(etree.tostring(pricing_query, **XML_OPTIONS))
        response = self._get_response(pricing_query, self.pricing_query_request.xml)

        # Check if any error is returned
        errors = response.element.xpath('//ns:error',
                                        namespaces={'ns': XHTML_NAMESPACE})
        if len(errors) > 0 and hasattr(errors[0], 'text'):
            for error in errors:
                product_reference = error.getprevious()
                # code="ERR_120"
                # Service Pricing : Product not found.
                if product_reference is not None:
                    logger.warning("EAN: {0}. {1}".format(product_reference.text,
                                                          error.text))
                # code="ERR_105"
                # The limit of requested elements is reached by the service.
                else:
                    logger.warning("{0}".format(error.text))

        return response

    def query_batch(self):
        """Return information about your currently processing import batches

        Usage::

            response = manager.query_batch()

        :returns: :class:`Response <Response>` object

        """
        batch_query = create_xml_element(self.connection, self.token, 'batch_query')
        self.batch_query_request = Request(etree.tostring(batch_query, **XML_OPTIONS))
        return self._get_response(batch_query, self.batch_query_request.xml)

    def query_carriers(self):
        """Return the available carriers managed on FNAC Marketplace platform

        Usage::

            response = manager.query_carriers()

        :returns: :class:`Response <Response>` object

        """
        carriers_query = create_xml_element(self.connection, self.token, 'carriers_query')
        etree.SubElement(carriers_query, "query").text = etree.CDATA("all")
        self.carriers_query_request = Request(etree.tostring(carriers_query, **XML_OPTIONS))
        return self._get_response(carriers_query, self.carriers_query_request.xml)

    def query_client_order_comments(self, results_count='', **elements):
        return self._query('client_order_comments', **elements)

    def update_client_order_comments(self, seller_comment, order_fnac_id):
        """Reply to client order comments

        Usage::

            response = manager.update_client_order_comments(seller_comment,
                                                            order_fnac_id)

        :type seller_comment: str
        :param seller_comment: The seller comment

        :type order_fnac_id: str
        :param order_fnac_id: Order unique identifier filter from FNAC

        :returns: :class:`Response <Response>` object

        """
        client_order_comments_update = create_xml_element(self.connection, self.token,
                                                          'client_order_comments_update')
        comment = etree.Element('comment', id=order_fnac_id)
        etree.SubElement(comment, 'comment_reply').text = etree.CDATA(seller_comment)
        client_order_comments_update.append(comment)
        self.client_order_comments_update_request = \
            Request(etree.tostring(client_order_comments_update, **XML_OPTIONS))
        return self._get_response(client_order_comments_update,
                                  self.client_order_comments_update_request.xml)

    def query_messages(self, results_count='', **elements):
        return self._query('messages', results_count, **elements)

    def update_messages(self, messages):
        """Update message sent on your offers or orders : reply, set as read, ...

        Usage::

            response = manager.update_messages(messages)

        :type messages: Message
        :param messages: the specified messages we want to update

        :returns: :class:`Response <Response>` object

        Example::

            >>> m1 = Message(action='mark_as_read', id='12345')
            >>> m2 = Message(action='reply', id='12345')
            >>> m2.description = 'Your order has been shipped'
            >>> m2.subject = 'order_information'
            >>> m2.type = 'ORDER'
            >>> response = manager.update_messages([m1, m2])

        """
        messages_update = create_xml_element(self.connection, self.token, 'messages_update')
        for m in messages:
            message = etree.XML(dict2xml(m.to_dict()))
            messages_update.append(message)

        self.messages_update_request = Request(etree.tostring(messages_update, **XML_OPTIONS))
        return self._get_response(messages_update, self.messages_update_request.xml)

    def query_incidents(self, results_count='', **elements):
        return self._query('incidents', results_count, **elements)

    def update_incidents(self, order_id, incident_update_action, reasons):
        """Handle incidents created on orders

        Usage::

            response = manager.update_incidents(order_id,
                                                incident_update_action,
                                                reasons)

        :type order_id: str
        :param order_id: the unique FNAC identified for an order

        :type incident_update_action: str
        :param incident_update_action: the action to perform (`'refund'` is the
                                       only available action for the moment)

        :type reasons: list
        :param reasons: the reasons of the incident for this order

        Example::

            reason = {"order_detail_id": 1, "refund_reason": 'no_stock'}
            response = manager.update_incidents('07LWQ6278YJUI', 'refund', [reason])

        :returns: :class:`Response <Response>` object

        """
        incidents_update = create_xml_element(self.connection, self.token,
                                              'incidents_update')
        order = etree.Element('order', order_id=order_id,
                              action=incident_update_action)

        for reason in reasons:
            order_detail = etree.Element("order_detail")
            etree.SubElement(order_detail, 'order_detail_id').text = str(reason['order_detail_id'])
            etree.SubElement(order_detail, 'refund_reason').text = str(reason['refund_reason'])
            order.append(order_detail)

        incidents_update.append(order)
        self.incidents_update_request = \
            Request(etree.tostring(incidents_update, **XML_OPTIONS))
        return self._get_response(incidents_update,
                                  self.incidents_update_request.xml)

    def query_shop_invoices(self, results_count='', **elements):
        return self._query('shop_invoices', results_count, **elements)


# Dynamically set the docstrings for some query methods
for query_type in FnapyManager.VALID_QUERY_TYPES:
    method = getattr(FnapyManager, 'query_' + query_type)
    if is_py3:
        method.__doc__ = _create_docstring(query_type)
    else:
        method.__func__.__doc__ = _create_docstring(query_type)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.

"""
Exceptions in fnapy
"""


class FnapyException(Exception):
    """Main exception class"""
    pass


class FnapyUpdateOfferError(FnapyException):
    """Raised when the update of an offer is not valid"""
    pass


class FnapyPricingError(FnapyException):
    """Raised when no prices is found when using pricing_query"""
    pass


class FnapyConnectionError(FnapyException):
    """Raised when the connection is incorrect"""
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.

"""
Useful constants and configs.
"""

XHTML_NAMESPACE = "http://www.fnac.com/schemas/mp-dialog.xsd"
HEADERS = {'Content-Type': 'text/xml'}
XML_OPTIONS = {'pretty_print': True, 'xml_declaration': True, 'encoding': 'utf-8'}


# The sub elements in the order element
ORDER_ELEMENTS = (
    'shop_id', 'client_id', 'client_firstname', 'client_lastname',
    'client_email', 'adherent_number', 'order_id', 'order_culture', 'state',
    'created_at', 'fees', 'nb_messages', 'vat_rate', 'delivery_note',
    'shipping_address', 'billing_address', 'order_detail'
)


class Parameter(object):
    def __init__(self, name, desc):
        self.name = name
        self.desc = desc

    def change_desc(self, new_desc):
        """Return a new parameter object with a new description

        :param new_desc: a new description

        Sometimes in the FNAC API, the name of a node is used with a different
        type and a different meaning. This method allows to deal with these
        inconsistencies. As we don't manage the types, we just need to change
        the description.

        """
        return Parameter(self.name, new_desc)


paging               = Parameter('paging', 'Page number to fetch')
date                 = Parameter('date', 'Date filter')
quantity             = Parameter('quantity', 'Quantity filter')
product_fnac_id      = Parameter('product_fnac_id', "Product's unique identifier from fnac")
offer_fnac_id        = Parameter('offer_fnac_id', 'Offer unique identifier from fnac')
offer_seller_id      = Parameter('offer_seller_id', 'Offer unique identifier from seller (SKU)')
sort_by              = Parameter('sort_by', 'Quantity filter')
state                = Parameter('state', 'Order state filter')
states               = Parameter('states', 'Multiple order state filter')
order_fnac_id        = Parameter('order_fnac_id', 'Order unique identifier filter')
orders_fnac_id       = Parameter('orders_fnac_id', 'Multiple order unique identifier filter')
rate                 = Parameter('rate', 'Order rate filter')
message              = Parameter('message', 'Messages to update')
message_type         = Parameter('message_type', 'Message type filter')
message_archived     = Parameter('message_archived', 'Message is archived or not filter')
message_state        = Parameter('message_state', 'Message state')
message_id           = Parameter('message_id', 'Message unique identifier from fnac filter')
message_from_types   = Parameter('message_from_types', 'Messages author filter')
status               = Parameter('status', 'Status of the incident')
type                 = Parameter('type', 'Incident order detail opening reason')
types                = Parameter('types', 'Define the opening reasons of the incidents to be retrieved.')
incident_id          = Parameter('incident_id', 'A unique incident id')
incidents_id         = Parameter('incidents_id', 'Can contain multiple incident_id node (up to 50 max)')
closed_statuses      = Parameter('closed_statuses', 'Can contain multiple closed_status node (up to 50 max)')
closed_status        = Parameter('closed_status', 'Incident closing reason')
opened_by            = Parameter('opened_by', 'Who opened the incident')
closed_by            = Parameter('closed_by', 'Who closed the incident')
order                = Parameter('order', 'Orders to update')
orders               = Parameter('orders', 'Can contains multiple order node (up to 50 max)')
page                 = Parameter('page', 'Page number')
total_paging         = Parameter('total_paging', 'Number of pages available')
nb_total_per_page    = Parameter('nb_total_per_page', 'Number of results per page')
nb_total_result      = Parameter('nb_total_result', 'Number of results')
batch_id             = Parameter('batch_id', 'Batch unique identifier from fnac')
offer                = Parameter('offer', 'Offers')
batch                = Parameter('batch', 'List of batches')
nb_batch_running     = Parameter('nb_batch_running', 'Number of batches being processed')
nb_batch_active      = Parameter('nb_batch_active', 'Number of batches waiting to be processed')
carrier              = Parameter('carrier', 'A carrier')
incident             = Parameter('incident', 'Incident information')
pricing_product      = Parameter('pricing_product', 'Pricings of product')
comment              = Parameter('comment', 'Updated comments')
shop_invoice         = Parameter('shop_invoice', 'Shop invoices')
waiting_for_seller_answer = Parameter('waiting_for_seller_answer',
                                      """Fnac customer service can ask the seller
                                      to deal specifically with an incident.""")
client_order_comment = Parameter('client_order_comment', 'Client order comments')
client_order_comment_id = Parameter('client_order_comment_id',
                                    'Order unique identifier filter from fnac')

product_reference = Parameter('product_reference', 'Product referenc')
offer_reference = Parameter('offer_reference', 'Offer reference')
price = Parameter('price', 'Offer price')
product_state = Parameter('product_state', 'Product state of offer')
description = Parameter('description', 'Product description')
internal_comment = Parameter('internal_comment',
                             'Offer internal comment for personal use')
showcase = Parameter('showcase', 'Offer position in shop’s showcase')
treatment = Parameter('treatment', 'Treatment to do on offer')
pictures = Parameter('pictures', 'Add pictures to offer')

# The elements for the requests
REQUEST_ELEMENTS = {}
REQUEST_ELEMENTS['offers_query'] = (
    paging, date, quantity, product_fnac_id, offer_fnac_id,
    offer_seller_id
)
REQUEST_ELEMENTS['offers_update'] = (
    product_reference, offer_reference, price, product_state,
    quantity.change_desc('Offer quantity'), description,
    internal_comment, showcase, treatment, pictures
)
REQUEST_ELEMENTS['orders_query'] = (
    paging, date, sort_by, product_fnac_id, offer_fnac_id,
    offer_seller_id, state, states, order_fnac_id, orders_fnac_id
)
REQUEST_ELEMENTS['client_order_comments_query'] = (
    paging, date, rate, client_order_comment_id, order_fnac_id
)
REQUEST_ELEMENTS['messages_query'] = (
    paging, date, message_type, message_archived, message_state,
    message_id, order_fnac_id, offer_fnac_id, offer_seller_id,
    sort_by, message_from_types
)
REQUEST_ELEMENTS['incidents_query'] = (
    paging, date, status, type, types, incident_id, incidents_id,
    closed_statuses, closed_status, waiting_for_seller_answer,
    opened_by, closed_by, sort_by,
    order.change_desc('A unique order id'),
    orders.change_desc('Can contains multiple order node (up to 50 max)')
)
REQUEST_ELEMENTS['shop_invoices_query'] = (
    paging, date
)

# The elements for the responses
RESPONSE_ELEMENTS = {}
RESPONSE_ELEMENTS['offers_update'] = (
    batch_id,
)
RESPONSE_ELEMENTS['offers_query'] = (
    page, total_paging, nb_total_per_page, nb_total_result, offer
)
RESPONSE_ELEMENTS['orders_update'] = (
    order,
)
RESPONSE_ELEMENTS['orders_query'] = (
    page, total_paging, nb_total_per_page, nb_total_result,
    order.change_desc('Orders')
)
RESPONSE_ELEMENTS['batch_status'] = (batch_id, offer)
RESPONSE_ELEMENTS['batch_query'] = (
    batch, nb_batch_running, nb_batch_active
)
RESPONSE_ELEMENTS['carriers_query'] = (carrier,)
RESPONSE_ELEMENTS['incidents_query'] = (
    page, total_paging,
    nb_total_per_page, nb_total_result, incident
)
RESPONSE_ELEMENTS['incidents_update'] = (
    order.change_desc('Seller Order in fnac marketplace'),
)
RESPONSE_ELEMENTS['messages_query'] = (
    page, total_paging,
    nb_total_per_page, nb_total_result,
    message.change_desc('Messages list')
)
RESPONSE_ELEMENTS['messages_update'] = (message,)
RESPONSE_ELEMENTS['pricing_query'] = (pricing_product, )
RESPONSE_ELEMENTS['client_order_comments_query'] = (
    page, total_paging, nb_total_per_page, nb_total_result, client_order_comment,
)
RESPONSE_ELEMENTS['client_order_comments_update'] = (comment,)
RESPONSE_ELEMENTS['shop_invoices_query'] = (
    page, total_paging, nb_total_per_page, nb_total_result, shop_invoice,
)



# -*- coding: utf-8 -*-
import sys

# Syntax sugar.
_ver = sys.version_info

#: Python 2.x?
is_py2 = (_ver[0] == 2)

#: Python 3.x?
is_py3 = (_ver[0] == 3)

#: Python 3.3.x
is_py33 = (is_py3 and _ver[1] == 3)

#: Python 3.4.x
is_py34 = (is_py3 and _ver[1] == 4)

#: Python 2.7.x
is_py27 = (is_py2 and _ver[1] == 7)


try:  # pragma: no cover
    import __pypy__
    is_pypy = True
except:  # pragma: no cover
    __pypy__ = None
    is_pypy = False


if is_py3:
    builtin_str = str
    str = str
    bytes = bytes
    basestring = (str, bytes)
    integer_types = (int, )
    numeric_types = integer_types + (float, )

    def iterkeys(d):
        return iter(d.keys())

    def itervalues(d):
        return iter(d.values())

    def iteritems(d):
        return iter(d.items())

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    def is_bytes(x):
        return isinstance(x, (bytes, memoryview, bytearray))

    from collections import Callable

    def callable(obj):
        return isinstance(obj, Callable)

else:
    builtin_str = str
    bytes = str
    str = unicode
    basestring = basestring
    integer_types = (int, long)
    numeric_types = integer_types + (float, )

    def iterkeys(d):
        return d.iterkeys()

    def itervalues(d):
        return d.itervalues()

    def iteritems(d):
        return d.iteritems()

    exec('def reraise(tp, value, tb=None):\n raise tp, value, tb')

    def is_bytes(x):
        return isinstance(x, (buffer, bytearray))

    callable = callable


def with_metaclass(meta, base=object):
    return meta("NewBase", (base,), {})


def to_unicode(obj, encoding='utf-8'):
    """
    Convert ``obj`` to unicode"""
    # unicode support
    if isinstance(obj, str):
        return obj

    # bytes support
    if is_bytes(obj):
        if hasattr(obj, 'tobytes'):
            return str(obj.tobytes(), encoding)
        return str(obj, encoding)

    # string support
    if isinstance(obj, basestring):
        if hasattr(obj, 'decode'):
            # ignoring utf8 errors, as fnac send us malformed
            # utf8 xml sometimes ...
            return obj.decode(encoding, errors='ignore')
        else:
            return str(obj, encoding)

    return str(obj)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.

"""
The class storing all the information for the connection to the FNAC API
"""

from fnapy.exceptions import FnapyConnectionError
from fnapy.utils import get_credentials


def check_credentials_validity(credentials):
    invalid_credentials = []
    for k, v in credentials.items():
        if v is None:
            invalid_credentials.append(k)

    if len(invalid_credentials):
        msg = 'These credentials are invalid: '
        msg += ', '.join(invalid_credentials)
        raise FnapyConnectionError(msg)


class FnapyConnection(object):
    """The connection class of fnapy

    Usage::
        connection = FnapyConnection(credentials=credentials, sandbox=sandbox)

    Example:

    * Create a connection with a credentials dictionary::

        credentials = {'partner_id': 'my_partner_id', 'shop_id': 'my_shop_id',
                       'key': 'my_key', 'sandbox': False}

    * Create a connection using the environment variables for a given account
      type (the sandbox account is used if sandbox True. If sandbox is False, the
      real account is used)::

        connection = FnapyConnection(sandbox=True)

    .. note:: You must have previously defined the following environment
        variables:

    * for the sandbox: FNAC_SANDBOX_PARTNER_ID, FNAC_SANDBOX_SHOP_ID,
        FNAC_SANDBOX_KEY

    * for the real account: FNAC_PARTNER_ID, FNAC_SHOP_ID, FNAC_KEY

    """
    def __init__(self, credentials={}, sandbox=None):
        # credentials
        if len(credentials) > 0:
            expecteds = ('partner_id', 'shop_id', 'key', 'sandbox')
            for expected in expecteds:
                if credentials.get(expected) is None:
                    msg = "You didn't provide the {}."
                    msg += 'You must provide the following keys in credentials: '
                    msg += ', '.join(expecteds)
                    raise FnapyConnectionError(msg.format(expected))

        # sandbox
        else:
            if sandbox is not None:
                credentials = get_credentials(sandbox)
                credentials.update({'sandbox': sandbox})
            else:
                msg = 'You must either specify credentials or sandbox as arguments.'
                raise FnapyConnectionError(msg)

        check_credentials_validity(credentials)
        self.partner_id = credentials['partner_id']
        self.shop_id = credentials['shop_id']
        self.key = credentials['key']
        self.sandbox = credentials['sandbox']



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2016  <>
#
# Distributed under terms of the MIT license.


import re
import os.path as op


from codecs import open
from setuptools import setup, find_packages


def read(fname):
    ''' Return the file content. '''
    here = op.abspath(op.dirname(__file__))
    with open(op.join(here, fname), 'r', 'utf-8') as fd:
        return fd.read()

readme = read('README.rst')
changelog = read('docs/CHANGELOG.rst').replace('.. _changelog:', '')

install_requirements = [
    'requests',
    'xmltodict',
    'lxml',
]

version = ''
version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                    read(op.join('fnapy', '__init__.py')),
                    re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')


setup(
    name='fnapy',
    author='Taurus Olson',
    author_email='taurusolson@gmail.com',
    version=version,
    description='A Python API for FNAC WebServices',
    long_description=readme + '\n\n' + changelog,
    keywords=['api', 'fnac', 'python', 'webservices'],
    packages=find_packages(),
    install_requires=install_requirements,
    zip_safe=True,
    license='MIT',
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
